{-# OPTIONS -cpp #-}
{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module Main where

import Prelude hiding		    ( catch )
import Control.Exception
import Control.Monad
import Control.Monad.Error	    ( catchError, throwError )
import Control.Monad.Trans	    ( liftIO )
import Language.Haskell.Pretty
import System			    ( exitFailure, getProgName, getArgs )
import Text.PrettyPrint

import Interface.Interactive
import Interface.Options            -- ( usage, ...)
import Interface.ChaseImports

import Plugins			    
import Plugins.All		    ()    -- To force compilation

import Syntax.ConcreteToInternal
import Syntax.ParseMonad
import Syntax.Parser
import qualified Syntax.Concrete as C

import TypeChecker		    ( checkProgram )
import TypeChecker.Monad
import TypeChecker.Monad.Options

import Utils.Error
import Utils.Monads

concatPrograms :: [ParseResult C.Program] -> TC C.Program
concatPrograms rs =
    do	z <- liftIO $ (Right <$> evaluate (concatPrograms' rs))
		`catch` \e ->
		    case e of
			ErrorCall err	-> return (Left err)
			_		-> throwIO e
	case z of
	    Left err	-> fail err -- TODO: Better error
	    Right p	-> return p
    where
	concatPrograms' [] = []
	concatPrograms' (r:rs) =
	    case r of
		ParseFailed p why what -> error $ showParseFailure p why what
		ParseOk _ prog ->
		    prog ++ concatPrograms' rs

parseFiles :: TC C.Program
parseFiles =
    do	opts <- commandLineOptions
	rs <- liftIO $ mapM parseProgram (optInputFiles opts)
	concatPrograms rs
    where
	parseProgram file
	    | isLiterate file	= parseLiterateFile programParser file
	    | otherwise		= parseFile programParser file

runAgdaLight :: TC ()
runAgdaLight =
    do	-- Parse (standard) options. Plugin options are parsed
	-- when loading the plugins.
	progName    <- liftIO $ getProgName
	argv	    <- liftIO $ getArgs
	let opts' = parseStandardOptions
			progName
			[ a | a <- argv, notPluginOption a ]
	case opts' of
	    Left err	-> liftIO $ optionError err
	    Right opts
		| optShowHelp opts	-> liftIO printUsage
		| optShowVersion opts	-> liftIO printVersion
		| null (optInputFiles opts)
					-> liftIO printUsage
		| length (optInputFiles opts) > 1
					-> liftIO multipleFilesDeprecated
		| otherwise		-> checkFiles opts
    where
	checkFiles opts =
	    do	files <- liftIO $ chaseImports	(optIncludeDirs opts)
						(optInputFiles opts)
		when (optVerbose opts >= 1) $ liftIO $
		    do	putStrLn "After import chasing:"
			mapM_ (\file -> putStrLn $ "  " ++ file) files
		setCommandLineOptions (opts { optInputFiles = files }) $
		    do  when (optInteractive opts) $
			    liftIO $ putStr splashScreen
		    

			(if optInteractive opts
				then interactionLoop
				else id ) $
			    do	-- Clear the state
				clearState

				-- Load plugins
				loadPlugins `catchError` \e ->
				    case errCode e of
					OptionError -> liftIO $ optionError
								    (show $ errWhy e)
					_	    -> throwError e

				-- Get options
				opts <- commandLineOptions
			    
				-- Parse input files
				prog <- parseFiles

				-- Convert to internal syntax
				let iprog = programToInternal prog

				-- Print parsed program
				when (optVerbose opts >= 1) $
				    liftIO $ print iprog

				-- Type check
				defs <- checkProgram iprog

				-- Let plugins finish up
				finalizePlugins

				return defs
			return ()

optionError :: String -> IO ()
optionError err =
    do	putStr $ "Bad arguments: " ++ err
	printUsage
	exitFailure

multipleFilesDeprecated :: IO ()
multipleFilesDeprecated =
    do	print $
	    text "NOTE:" $$ nest 2 (vcat
		[ fsep $ pwords $
		    "The passing of multiple files to agdaLight is " ++
		    "no longer supported. To achieve the same effect " ++
		    "use the new import pragmas:"
		, nest 2 $ text "-- # import Property"
		, nest 2 $ text "-- # import QuickCheck"
		, nest 2 $ text "-- # import examples.fol.List"
		, fsep $ pwords $
		    "Imports are looked for in"
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "the same directory as the file " ++
					    "containing the import,"
				)
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "the current working directory,")
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "any directory specified using the -i flag")
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "the agdaLight lib directory (" ++
					    _ROOT_ ++ "/lib) directory.")
		, fsep $ pwords $ "To find a file the dots are replaced by slashes and '.agda'" ++
				    " is appended to the module name, so examples.fol.List" ++
				    " becomes examples/fol/List.agda."
		]
	    )
	exitFailure

printVersion :: IO ()
printVersion =
    do	putStrLn $ "AgdaLight version " ++ _VERSION_
	putStrLn "  Andreas Abel, Thierry Coquand, Ulf Norell"

printUsage :: IO ()
printUsage =
    do	progName <- getProgName
	putStr $ usage standardOptions_ allPluginInfo progName

main :: IO ()
main =
    do	r <- runTC runAgdaLight
	case r of
	    Left err	-> do	print err
				exitFailure
	    Right ()	-> return ()

