
module Plugins.FOL.Types where

import Control.Monad.State
import Text.PrettyPrint

import Syntax.Internal
import Syntax.Pretty

import Plugins.FOL.Constants

import Utils.Names
import Utils.Map

-- FOL propositions -------------------------------------------------------

data Formula = Predicate String [FOLTerm]
	     | And Formula Formula
	     | Or Formula Formula
	     | Not Formula
             | Implies Formula Formula
             | Equiv Formula Formula
	     | ForAll String (FOLTerm -> Formula)
	     | Exists String (FOLTerm -> Formula)
	     | TRUE
	     | FALSE

data FOLTerm	= FOLFun String [FOLTerm]
		| FOLVar String

-- | A simple textual representation of counter model
type CounterModel = [String]

implies :: Formula -> Formula -> Formula
implies = Implies
--implies p q = Not p `Or` q

equiv :: Formula -> Formula -> Formula
equiv = Equiv
--equiv p q   = (p `implies` q) `And` (q `implies` p)

-- | > equal = Predicate "equal"
--
--   This refers to the predefined equality in otter.
equal :: [FOLTerm] -> Formula
equal	= Predicate kwEqual
kwEqual = "equal"

-- | > app = FOLFun "app_"
--
--   This is just an arbitrary constant which we hope doesn't
--   conflict with user constants.
app :: [FOLTerm] -> FOLTerm
app	= FOLFun kwApp
kwApp = "app_"

fol_propExtensionality = "__fol_extensionality"
propExtensionality =
    ForAll "f" $ \f ->
    ForAll "g" $ \g ->
	(ForAll "x" $ \x -> equal [app [f,x], app [g,x]])
	`implies`
	equal [f, g]

fol_propPairElimination = "__fol_pair_elimination"
propPairElimination =
    (	ForAll "x" $ \x ->
	ForAll "y" $ \y ->
	    equal [ proj [ pair [x, y], FOLFun kwFst [] ], x ]
		`And`
	    equal [ proj [ pair [x, y], FOLFun kwSnd [] ], y ]
    )

fol_propPairInjectivity = "__fol_pair_injectivity"
propPairInjectivity =
    (	ForAll "x" $ \x ->
	ForAll "y" $ \y ->
	ForAll "z" $ \z ->
	ForAll "w" $ \w ->
	    equal [ pair [x,y], pair [z,w] ]
	    `Implies`
	    (equal [x,z] `And` equal [y,w])
    )

fol_primitives =
    [ (fol_propPairInjectivity, propPairInjectivity)
    , (fol_propPairElimination, propPairElimination)
    , (fol_propExtensionality, propExtensionality)
    ]

pair :: [FOLTerm] -> FOLTerm
pair = FOLFun kwPair
kwPair = "pair_"

proj :: [FOLTerm] -> FOLTerm
proj = FOLFun kwProj
kwProj = "proj_"

kwFst = "fst_"
kwSnd = "snd_"
--kwField x = "field_" ++ x ++ "_"

sigma :: [FOLTerm] -> FOLTerm
sigma = FOLFun kwSigma

kwSigma = "sigma_"

fun :: [FOLTerm] -> FOLTerm
fun = FOLFun kwPi

kwPi = "pi_"

-- | @bigAnd@ produces a big conjunction.
bigAnd :: [Formula] -> Formula
bigAnd [] = TRUE
bigAnd ps = foldr1 And ps

-- | The representation of a first order goal.
data FOLGoal = FOLGoal Name ([Formula],[Formula]) Formula [Value] Value
		-- ^ A goal consists of the position from which it was called,
		--   a list of premises (definitions of lambda lifted functions,
		--   and axioms), and the formula to be proved. We also
		--   keep the types for pretty printing reasons.

-- Pretty printing FOL propositions ---------------------------------------

{- using Constants instead
keywords = ["True", "False", "/\\", "\\/", "Not", "==>", "<=>", 
            "Forall", "Exists"]
[kwTrue, kwFalse, kwAnd, kwOr, kwNot, kwImplies, kwEquiv, 
 kwForall, kwExists] = keywords
-}

initialNames :: UsedNames
initialNames = foldl (\ m x -> updateMap x 0 m) noNamesUsed folConstants 
  --keywords

{- Precedence levels:

12 forall
11 exists
10 application right
 9 application left
 7 not
 6 equal
 4 and
 3 or
 2 implies
 1 equiv
-}

precedences :: [Int]
precedences = [0,1,2,3,4,6,7,9,10,11,12]

[precTop,precEquiv,precImplies,precOr,precAnd,precEqual,precNot,
 precAppL,precAppR,precExists,precForAll] = precedences
precMax = precAppR

-- indentation is 2 (nest)

instance Pretty FOLTerm where
    prettyPrec n (FOLVar x) = text x
    prettyPrec n (FOLFun x []) = text x
    prettyPrec n (FOLFun xProj [t,t']) | xProj == kwProj =
	    prettyPrec precMax t <> text "." <> prettyProj t'
	where
	    prettyProj (FOLFun p [])
		| p == kwFst	= text "1"
		| p == kwSnd	= text "2"
    prettyPrec n (FOLFun xApp [t,t']) | xApp == kwApp = 
        mparens (n>precAppL) $ 
                fsep [ prettyPrec precAppL t
                     , nest 2 $ prettyPrec precAppR t'
                     ]
    prettyPrec n (FOLFun x ts) = 
        mparens (n>precAppL) $ 
                fsep $ (text x) : map (nest 2 . prettyPrec precAppR) ts

-- TODO : Lamport-style of printing formulas
-- Nested Forall/Exists needs better layout

instance Pretty Formula where
    prettyPrec n p = flip evalState noNamesUsed $ prettyFM n p -- hack

prettyFM :: Int -> Formula -> State UsedNames Doc
prettyFM n TRUE      = return $ text fol_True
prettyFM n FALSE     = return $ text fol_False
prettyFM n (And p q) = do dp <- prettyFM precAnd p
                          dq <- prettyFM precAnd q
                          return $ mparens (n>precAnd) $
                            fsep [ dp
                                 , nest 2 $ text fol_And
                                 , nest 2 dq
                                 ]
prettyFM n (Or  p q) = do dp <- prettyFM precOr p
                          dq <- prettyFM precOr q
                          return $ mparens (n>precOr) $
                            fsep [ dp
                                 , nest 2 $ text fol_Or
                                 , nest 2 dq
                                 ]
prettyFM n (Not p)   = do dp <- prettyFM precNot p
                          return $ mparens (n>precNot) $
                            fsep [ text fol_Not
                                 , nest 2 $ dp
                                 ]
prettyFM n (Implies  p q) = 
    do dp <- prettyFM precImplies p
       dq <- prettyFM precImplies q
       return $ mparens (n>precImplies) $
                            fsep [ dp
                                 , nest 2 $ text fol_Implies
                                 , nest 2 dq
                                 ]
prettyFM n (Equiv p q) = 
    do dp <- prettyFM precEquiv p
       dq <- prettyFM precEquiv q
       return $ mparens (n>precEquiv) $
                            fsep [ dp
                                 , nest 2 $ text fol_Equiv
                                 , nest 2 dq
                                 ]
prettyFM n (ForAll x p) = 
    do name <- fresh x
       dp <- prettyFM precForAll (p (FOLVar name))
       return $ fsep $ (if (n==precForAll) then [] else [text fol_ForAll]) ++
                  [ nest 2 $ text $ name
                  , nest 2 $ dp
                  ]
prettyFM n (Exists x p) = 
    do name <- fresh x
       dp <- prettyFM precExists (p (FOLVar name))
       return $ fsep $ (if (n==precExists) then [] else [text fol_Exists]) ++
                  [ nest 2 $ text $ name
                  , nest 2 $ dp
                  ]
{-
prettyFM n (Exists x p) = do name <- fresh x
                             dp <- prettyFM precExists (p (FOLVar name))
                             return $ fsep [ text fol_Exists
                                           , nest 2 $ text $ name
                                           , nest 2 $ dp
                                           ]
-}
prettyFM n (Predicate xEqual [t,t']) | xEqual == kwEqual = 
    return $ mparens (n>precEqual) $ 
                fsep [ prettyPrec precTop t
                     , text "=="
                     , prettyPrec precTop t'
                     ]
prettyFM n (Predicate p ts) = do let dts = map (prettyPrec precAppR) ts
                                 return $ mparens (n>precAppL) $
                                        fsep $ (text p) :
                                               map (nest 2) dts

instance Show FOLTerm where
    show = show . pretty

instance Show Formula where
    show = show . pretty

