{-| Bidirectional type checking and checking of a 
    program (= list of definitions).
-}

module TypeChecker where

import Control.Monad		( when, unless )
import Control.Monad.Error
import Data.Maybe               ( isJust, fromJust )

import Plugins			( pluginCheck, pluginInfer )

import Syntax.Internal
import Syntax.Pretty

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Context
import TypeChecker.Monad.Options
import TypeChecker.Equality
import TypeChecker.Computation
import TypeChecker.Values
import TypeChecker.UnfoldConstants
import TypeChecker.Universes
import TypeChecker.Context
import TypeChecker.Errors
import TypeChecker.Terms

import Utils.Map
import Utils.Names		( fresh )
import Utils.Monads

-- Check program ----------------------------------------------------------

checkProgram :: Program -> TC ()
checkProgram [] = return ()
checkProgram (def : prog) =
    do	case def of
	    NormalDef name mt e -> checkDef name mt (Just e)
	    NewAxiom  name t	-> checkDef name (Just t) Nothing
	checkProgram prog

checkDef :: Name -> Maybe Term -> Maybe Term -> TC ()
checkDef name mt me =
    sayWhen (whenCheckingDefinition name) $

	do  defs <- getDefinitions

	    when (isInMap name defs) $ failure $ repeatedVariableError name

	    case (mt, me) of
		(Nothing, Just e) ->
		    do	v <- evaluateTypedTerm =<< inferType e
			addDefinition name (Definition v)
		(Just t, me)	->
		    do	-- Check that the type is well formed
			t' <- inferType t
			isType t'
			v <- evaluateTypedTerm t'

			case me of
			    Nothing -> addDefinition name (Axiom v)
			    Just e  ->
				do  v <- evaluateTypedTerm =<< checkType e v
				    addDefinition name (Definition v)

-- Type inference ---------------------------------------------------------

inferType :: Term -> TC TypedTerm
inferType e =
	sayWhen (whenInferringType e) $
	inferType' e
    where

	inferType' (Var x _) = Var x <$> (valueType <$> lookupVar x)
	inferType' (Def x _) = Def x <$> lookupDef x

	inferType' (App e1 e2) =
	    do	e1' <- inferType e1
		v1  <- termType e1'
		case unfoldConstants v1 of
		    VPi v _ _	-> do	e2' <- checkType e2 v
					return $ App e1' e2'
		    _		-> failure $ notAFunctionError e1 v1

	inferType' (Star u) = return $ Star u
	inferType' Prop	    = return Prop

	inferType' (Pi x a b _) =
	    do	a' <- inferType a
		sa <- isType a'
		v  <- evaluateTypedTerm a'
		extendContext x (Abstract v) $
		    do	b' <- inferType b
			sb <- isType b'
			return $ Pi x a' b' (VSort $ piSort sa sb)
	    where
		piSort (VSUniv 0) VSProp    = VSProp
		piSort sa sb		    = max sa sb

	inferType' (Let defs e) = typeLet defs $ inferType e
	inferType' (CallPlugin name es _) =
	    CallPlugin name es <$> pluginInfer name es

	inferType' e = failure $ cannotInferError e

-- Type checking ----------------------------------------------------------

-- | invariant: the 'Type' should be a type value and
-- the context should consist of (isType)s. If
-- @e' <- checkType e v@ then @termType e'@ is a subtype of @v@.
checkType :: Term -> Type -> TC TypedTerm
checkType t v =
	sayWhen (whenCheckingType t v) $
	    checkType' t v
    where

	checkType' (Lam x e _) v =
	    case unfoldConstants v of
		VPi w f _ ->
			do  e' <- extendContext x (Abstract w) $
				    checkType e $ f `clApp` VId x w
			    return $ Lam x e' v
		_ -> failure $ lambdaHasFunctionTypeError (Lam x e ()) v

	checkType' (CallPlugin name es _) v =
	    CallPlugin name es <$> pluginCheck name es v

-- Lets can be both inferred and checked, depending on whether the
-- body is inferrable or not.
	checkType' (Let defs e) v = typeLet defs $ checkType e v
	checkType' e v =
	    do	e' <- inferType e
		w <- termType e'
		sayWhat (checkingSubtypingConstraint w v) $
		    w =<= v
		return e'

-- | The normal let rule. The body of a local definition is known whenever
--   the definition is in scope.
typeLet :: [LocalDefinition] -> TC TypedTerm -> TC TypedTerm
typeLet [] k = Let [] <$> k
typeLet (LocalDef x Nothing b _ : defs) k =
    do	b' <- inferType b
	v  <- evaluateTypedTerm b'
	Let ds e <- extendContext x (Defined v) $ typeLet defs k
	return $ Let (LocalDef x Nothing b' v : ds) e
typeLet (LocalDef x (Just t) b _ : defs) k =
    do	t' <- inferType t
	isType t'
	b' <- checkType b =<< evaluateTypedTerm t'
	w  <- evaluateTypedTerm b'
	Let ds e <- extendContext x (Defined w) $ typeLet defs k
	return $ Let (LocalDef x (Just t') b' w : ds) e

-- Is type ----------------------------------------------------------------

isType :: TypedTerm -> TC SortValue
isType e =
    do	v <- termType e
	case unfoldConstants v of
	    VSort s -> return s
	    _	    -> failure $ notATypeError e v

-- Is sort ----------------------------------------------------------------

isSort :: TypedTerm -> TC SortValue
isSort e =
    do	v <- evaluateTypedTerm e
	case unfoldConstants v of
	    VSort s -> return s
	    _	    -> failure $ notASortError e v

