
module Plugins.Assert where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Text.PrettyPrint
import Control.Monad	    ( when )
import Control.Monad.Trans  ( liftIO )

import Interface.Options    ( Options(..) )

import Syntax.Name	    ( Name(..) )
import Syntax.Internal	    ( Term(..) )
import Syntax.Values	    ( Value, Type(..), vTop, typeMapM )
import Syntax.Position	    ( Range )
import Syntax.Pretty	    ( Pretty(..) )

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.UnfoldConstants
import TypeChecker.Values.DeBruijn
import TypeChecker.Values.Refresh
import TypeChecker.MetaVars

import Utils.Error
import Utils.Monads

import Plugins

type AssertOptions = ()
type AssertState   = [Assertion]

data Assertion = Assertion Range Type Type

-- TODO: monadify pretty printing
instance Pretty Assertion where
    pretty (Assertion r v u) = 
        sep [ pretty r <> text ":"
            , text "  ? ::" <+> pretty v
	    , text "    ==" <+> pretty u
            ]

instance Show Assertion where
    show = show . pretty

instance Refresh Assertion where
    refresh (Assertion r v w) = Assertion r <$> refresh v <*> refresh w

assertCheck :: AssertOptions -> AssertState -> Range -> [Term] -> Type ->
               TC (Value, AssertState)
assertCheck o s r []    v		= do v' <- typeMapM unBruijn v
					     u  <- typeMapM unfoldAllConstants v
					     u' <- typeMapM unBruijn u
					     return (vTop, s ++ [Assertion r v' u'])
assertCheck o s r [Var _ (Name _ "silent" _)] v	= return (vTop, s)
assertCheck o s r [Def _ (Name _ "silent" _)] v	= return (vTop, s)
assertCheck o s r [Con _ (Name _ "silent" _)] v	= return (vTop, s)
assertCheck o s r [Data _ (Name _ "silent" _)] v	= return (vTop, s)
assertCheck o s r (_:_) v		= failure assertWithArgumentsError 

assertFinal :: AssertOptions -> AssertState -> TC ()
assertFinal o [] = return ()
assertFinal o s = 
    do opts <- commandLineOptions 
       when (optVerbose opts >= 1) $
            do	liftIO $ print $ hcat $ [ text "=== Assertions ",
					text $ replicate 55 '='
				      ]
		mapM_ (liftIO . print <.> refresh) s

assertHelp :: String
assertHelp =
    unlines [ "Asserts existence of object."
	    , "  assert-plugin()       -- prints the asserted type if verbosity > 0"
	    , "  assert-plugin(silent) -- doesn't print asserted type"
	    ]

plugin :: SomePlugin
plugin = mkPlugin
    Plugin  { plName		= "assert"
            , plHelp            = assertHelp
	    , plInit		= \_ -> return []
	    , plFinal		= assertFinal
	    , plCheck		= assertCheck
	    , plInfer		= \_ _ _ _ -> failure noInferenceError
	    , plOptions		= simpleOptions []
	    , plDefaultOptions	= ()
	    }

noInferenceError :: ErrorFun
noInferenceError e =
    e	{ errWhat   = fsep $ pwords "type inference failed"
	, errWhy    = fsep $ pwords "the type of an assertion cannot be inferred"
	}

assertWithArgumentsError :: ErrorFun
assertWithArgumentsError e =
    e { errWhat = fsep $ pwords "assertion failed"
      , errWhy  = fsep $ pwords "assert does not accept arguments"
      }
