module TypeChecker.Values where

import Control.Monad
import Data.List

import Syntax.Internal
import Syntax.Values
import Syntax.Pretty
import Syntax.Position
import Syntax.Sorts
import Syntax.Values

import Utils.Map
import Utils.Pairs
import Utils.Debug
import Utils.Names

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Errors
import TypeChecker.Abstractions
import TypeChecker.Universes
import TypeChecker.Values.Refresh
import TypeChecker.Monad.Context
import TypeChecker.Context
import {-# SOURCE #-} TypeChecker.UnfoldConstants -- REMOVE_IF_HADDOCK

import Utils.Monads

{- | Give an argument to a function type.

>  piApp ((x :: A) -> B) w = B[w/x]
 
Raises an internal error if the first argument does not have a 
function type. -}
piApp :: Type -> Value -> TC Type
piApp t w =
    do	v' <- unfoldConstants v
	case vView v' of
	    VPi _ f -> return $ absApp f w
	    _	    -> failure $ customError $
				"Internal error in 'piApp': "
				++ show v
				++ " should be a function type but is "
				++ show v'
				++ " in the application to "
				++ show w
    where
	v = unEl t

piApps :: Type -> [Value] -> TC Type
piApps = foldM piApp

vApps :: Value -> [Value] -> Value
vApps = foldl vApply

vApply :: Value -> Value -> Value
vApply v w =
    case vView v of
	VLam f  -> absApp f w
	VTop    -> vTop
	_	-> vApp v w

-- | Create a non dependent function type.
vFun :: Type -> Type -> Type
vFun t1 t2 = el (piSort (getSort t1) (getSort t2))
		(vPi t1 (mkAbs "_" $ raise t2))

-- | Takes a value and its type (which is a sigma type) and a projection, and
--   returns the type of the projected component of the value.
sigmaProj :: Value -> Type -> ProjectionValue -> TC Type
sigmaProj v t p =
    do	t' <- unfoldConstants (unEl t)
	let (w, f) = case vView t' of
		VSigma w f  -> (w,f)
		_	    ->
		    error $ show $
			text "sigmaProj:" <+>
			    vcat [ hsep [ text "v =", pretty v ]
				 , hsep [ text "t =", pretty t' ]
				 , hsep [ text "p =", pretty p ]
				 ]
	return $
	    case p of
		VFst	    -> w
		VSnd	    -> absApp f (vProject v VFst)
		VField x    -> error $ "impossible? TypeChecker.Values.sigmaProj _ _ (VField " ++ show x ++ ")"

vProject :: Value -> ProjectionValue -> Value
vProject v p = vp (vView v) p
    where
	vp VTop	 p		    = vTop
	vp (VPair u v) VFst	    = u
	vp (VPair u v) VSnd	    = v
	vp (VStruct xs) (VField x)  =
	    case Prelude.lookup x xs of
		Nothing	-> error $ "vProject (" ++ show v ++ ") " ++ show x
		Just v	-> v
	vp _ p			    = v `vProj` p

vProjs :: Value -> [ProjectionValue] -> Value
vProjs = foldl vProject

vAppOrProjs :: Value -> [AppOrProj] -> Value
vAppOrProjs = foldl vAppOrProj

vAppOrProj :: Value -> AppOrProj -> Value
vAppOrProj v (IsApp w)	= vApply v w
vAppOrProj v (IsProj p)	= vProject v p

-- | Create a Pi from a list of types.
--   @piAbstraction [v1,..,vn] v = (x1:v1) -> .. -> (xn:vn) -> v@. 
--   @v@ should be valid in the context @..,v1,..,vn@ and @vi@ in the
--   context @..,v1,..,vi-1@. The result is valid in the same context as @v1@.
piAbstraction :: [Type] -> Type -> Type
piAbstraction ws v = foldr mkPi v ws
    where
	mkPi w b = el (piSort s1 s2) $ vPi w (mkAbs "x" b)
	    where
		s1 = getSort w
		s2 = getSort b

-- | Create @n@ lambdas.
lamAbstraction :: Int -> Value -> Value
lamAbstraction 0 v = v
lamAbstraction n v = vLam (mkAbs "x" $ lamAbstraction (n - 1) v)

-- | @endsIn ((x1::A1) -> .. -> (xn::An) -> B) k@
--   Returns (n, B). Doesn't bring anything into scope.
endsIn :: Type -> TC (Int, Type)
endsIn t = endsIn' 0 t
    where
	endsIn' n t =
	    do	v  <- unfoldConstants (unEl t)
		case vView v of
		    VPi _ f ->
			endsIn' (n + 1) (absBody f)
			    -- =<< raiseBy (-1) (clBody f)	    -- TODO: why raise by -1?
		    _ -> return (n, t)

-- | Apply a function to all sub terms of a value (bottom up). Doesn't go
--   inside definition values (would lead to non-termination). The context
--   is extended with names but not with types when going under binders.
applyEverywhere :: (Value -> TC Value) -> Value -> TC Value
applyEverywhere f v' =
    do	v  <- refreshTop v'
	case vView v of
	    VId _	-> f v
	    VBId _	-> f v
	    VDef _ _	-> f v
	    VCon _ _ _	-> f v
	    VSort _	-> f v
	    VMeta _	-> f v
	    VTop	-> f v
	    VApp v w	-> f =<< vApp <$> r v <*> r w
	    VPi v g	-> f =<< vPi <$> rT v <*> rTA g
	    VLam g	-> f =<< vLam <$> rVA g
	    VPair v w	-> f =<< vPair <$> r v <*> r w
	    VProj v p	-> f =<< vProj <$> r v <*> return p
	    VSigma v g	-> f =<< vSigma <$> rT v <*> rTA g
	    VStruct xs	-> f =<< vStruct <$> mapM rSt xs
    where
	r = applyEverywhere f
	rT t = typeMapM r t
	rTA f = rA rT f
	rVA f = rA r f

	rSt (x,v) = (,) x <$> r v

	rA r f = absMapM (\v ->
		    do	x <- fresh $ absName f
			extendContext x (Abstract oops) (r v)
		 ) f
	
	oops = error $ "applyEverywhere doesn't know what type to extend the context with"

-- | Smash a value. The context is extended with names but not with types when
--   going under binders.
foldValue :: ([a] -> a) -> (Value -> TC a) -> Value -> TC a
foldValue h f v' =
    do	v0  <- refreshTop v'
	case vView v0 of
	    VId _	-> f v0
	    VBId _	-> f v0
	    VDef _ _	-> f v0
	    VCon _ _ _	-> f v0
	    VSort _	-> f v0
	    VMeta _	-> f v0
	    VTop	-> f v0
	    VApp v w	-> h' [f v0, r v, r w]
	    VPi v g	-> h' [f v0, rT v, rTA g]
	    VLam g	-> h' [f v0, rVA g]
	    VPair v w	-> h' [f v0, r v, r w]
	    VProj v p	-> h' [f v0, r v]
	    VSigma v g	-> h' [f v0, rT v, rTA g]
	    VStruct xs	-> h' $ Prelude.map (r . snd) xs
    where
	h' xs = h <$> sequence xs
	r = foldValue h f
	rT t = r $ unEl t
	rTA f = rA rT f
	rVA f = rA r f

	rA r f =
	    do	x <- fresh $ absName f
		extendContext x (Abstract oops) $ r (absBody f)
	
	oops = error $ "applyEverywhere doesn't know what type to extend the context with"
